'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var biggystring = require('biggystring');
var jsonschema = require('jsonschema');
var sprintfJs = require('sprintf-js');
var uriJs = require('uri-js');
var bn_js = require('bn.js');

/* global */
const imageServerUrl = 'https://developer.airbitz.co/content';

const otherSettings = {
  etherscanApiServers: ['https://api.etherscan.io'],
  blockcypherApiServers: ['https://api.blockcypher.com'],
  superethServers: ['https://supereth1.edgesecure.co:8443']
};

const defaultSettings = {
  customFeeSettings: ['gasLimit', 'gasPrice'],
  otherSettings
};

const currencyInfo = {
  // Basic currency information:
  currencyCode: 'ETH',
  currencyName: 'Ethereum',
  pluginName: 'ethereum',
  walletTypes: ['wallet:ethereum'],

  defaultSettings,

  addressExplorer: 'https://etherscan.io/address/%s',
  transactionExplorer: 'https://etherscan.io/tx/%s',

  denominations: [
  // An array of Objects of the possible denominations for this currency
  {
    name: 'ETH',
    multiplier: '1000000000000000000',
    symbol: 'Ξ'
  }, {
    name: 'mETH',
    multiplier: '1000000000000000',
    symbol: 'mΞ'
  }],
  symbolImage: `${imageServerUrl}/ethereum-logo-solo-64.png`,
  symbolImageDarkMono: `${imageServerUrl}/ethereum-logo-solo-64.png`,
  metaTokens: [
  // Array of objects describing the supported metatokens
  {
    currencyCode: 'REP',
    currencyName: 'Augur',
    denominations: [{
      name: 'REP',
      multiplier: '1000000000000000000'
    }],
    contractAddress: '0x1985365e9f78359a9B6AD760e32412f4a445E862',
    symbolImage: `${imageServerUrl}/augur-logo-solo-64.png`
  }, {
    currencyCode: 'DAI',
    currencyName: 'Dai Stablecoin',
    denominations: [{
      name: 'DAI',
      multiplier: '1000000000000000000'
    }],
    contractAddress: '0x89d24a6b4ccb1b6faa2625fe562bdd9a23260359',
    symbolImage: `${imageServerUrl}/dai-logo-solo-64.png`
  }, {
    currencyCode: 'WINGS',
    currencyName: 'Wings',
    denominations: [{
      name: 'WINGS',
      multiplier: '1000000000000000000'
    }],
    contractAddress: '0x667088b212ce3d06a1b553a7221E1fD19000d9aF',
    symbolImage: `${imageServerUrl}/wings-logo-solo-64.png`
  }, {
    currencyCode: 'USDT',
    currencyName: 'Tether',
    denominations: [{
      name: 'USDT',
      multiplier: '1000000'
    }],
    contractAddress: '0xdac17f958d2ee523a2206206994597c13d831ec7',
    symbolImage: `${imageServerUrl}/tether-logo-solo-64.png`
  }, {
    currencyCode: 'IND',
    currencyName: 'Indorse',
    denominations: [{
      name: 'IND',
      multiplier: '1000000000000000000'
    }],
    contractAddress: '0xf8e386EDa857484f5a12e4B5DAa9984E06E73705',
    symbolImage: `${imageServerUrl}/indorse-logo-solo-64.png`
  }, {
    currencyCode: 'HUR',
    currencyName: 'Hurify',
    denominations: [{
      name: 'HUR',
      multiplier: '1000000000000000000'
    }],
    contractAddress: '0xCDB7eCFd3403Eef3882c65B761ef9B5054890a47',
    symbolImage: `${imageServerUrl}/hur-logo-solo-64.png`
  }]
};

/**
 * Created by paul on 8/26/17.
 * 
 */

const Buffer$2 = require('buffer/').Buffer;

function normalizeAddress(address) {
  return address.toLowerCase().replace('0x', '');
}

function addHexPrefix(value) {
  if (value.startsWith('0x')) {
    return value;
  } else {
    return '0x' + value;
  }
}

function validateObject(object, schema) {
  const result = jsonschema.validate(object, schema);

  if (result.errors.length === 0) {
    return true;
  } else {
    for (const n in result.errors) {
      const errMsg = result.errors[n].message;
      console.log('ERROR: validateObject:' + errMsg);
    }
    return false;
  }
}

function bufToHex(buf) {
  const signedTxBuf = Buffer$2.from(buf);
  const hex = '0x' + signedTxBuf.toString('hex');
  return hex;
}

function toHex(num) {
  return biggystring.bns.add(num, '0', 16);
}

function isHex(h) {
  const out = /^[0-9A-F]+$/i.test(h);
  return out;
}

/**
 * Created by paul on 8/26/17.
 * 
 */

const ES_FEE_LOW = 'low';
const ES_FEE_STANDARD = 'standard';
const ES_FEE_HIGH = 'high';
const ES_FEE_CUSTOM = 'custom';

function calcMiningFee(spendInfo, networkFees) {
  if (spendInfo.spendTargets && spendInfo.spendTargets.length && spendInfo.spendTargets[0].publicAddress) {
    const { customNetworkFee } = spendInfo || {};
    if (spendInfo.networkFeeOption === ES_FEE_CUSTOM && customNetworkFee) {
      const { gasLimit, gasPrice } = customNetworkFee;
      const gasPriceGwei = biggystring.bns.mul(gasPrice, '1000000000');
      if (gasLimit && biggystring.bns.gt(gasLimit, '0') && gasPrice && biggystring.bns.gt(gasPrice, '0')) {
        return { gasLimit, gasPrice: gasPriceGwei };
      }
    }
    const targetAddress = normalizeAddress(spendInfo.spendTargets[0].publicAddress);
    let networkFeeForGasPrice = networkFees['default'];
    let networkFeeForGasLimit = networkFees['default'];

    if (typeof networkFees[targetAddress] !== 'undefined') {
      networkFeeForGasLimit = networkFees[targetAddress];
      if (typeof networkFeeForGasLimit.gasPrice !== 'undefined') {
        networkFeeForGasPrice = networkFeeForGasLimit;
      }
    }

    let useLimit = 'regularTransaction';
    if (spendInfo.currencyCode && spendInfo.currencyCode !== 'ETH') {
      useLimit = 'tokenTransaction';
    }

    let networkFeeOption = 'standard';
    if (typeof spendInfo.networkFeeOption === 'string') {
      networkFeeOption = spendInfo.networkFeeOption;
    }

    const gasLimit = networkFeeForGasLimit.gasLimit[useLimit];
    let gasPrice = '';
    if (!spendInfo.spendTargets[0].nativeAmount) {
      throw new Error('ErrorInvalidNativeAmount');
    }
    let nativeAmount = spendInfo.spendTargets[0].nativeAmount;
    if (useLimit === 'tokenTransaction') {
      // Small hack. Edgetimate the relative value of token to ethereum as 10%
      nativeAmount = biggystring.bns.div(nativeAmount, '10');
    }
    if (!networkFeeForGasPrice.gasPrice) {
      throw new Error('ErrorInvalidGasPrice');
    }
    const gasPriceObj = networkFeeForGasPrice.gasPrice;
    switch (networkFeeOption) {
      case ES_FEE_LOW:
        gasPrice = gasPriceObj.lowFee;
        break;
      case ES_FEE_STANDARD:
        if (biggystring.bns.gte(nativeAmount, networkFeeForGasPrice.gasPrice.standardFeeHighAmount)) {
          gasPrice = gasPriceObj.standardFeeHigh;
          break;
        }
        if (biggystring.bns.lte(nativeAmount, gasPriceObj.standardFeeLowAmount)) {
          if (!networkFeeForGasPrice.gasPrice) {
            throw new Error('ErrorInvalidGasPrice');
          }
          gasPrice = networkFeeForGasPrice.gasPrice.standardFeeLow;
          break;
        }

        // Scale the fee by the amount the user is sending scaled between standardFeeLowAmount and standardFeeHighAmount
        const lowHighAmountDiff = biggystring.bns.sub(gasPriceObj.standardFeeHighAmount, gasPriceObj.standardFeeLowAmount);
        const lowHighFeeDiff = biggystring.bns.sub(gasPriceObj.standardFeeHigh, gasPriceObj.standardFeeLow);

        // How much above the lowFeeAmount is the user sending
        const amountDiffFromLow = biggystring.bns.sub(nativeAmount, gasPriceObj.standardFeeLowAmount);

        // Add this much to the low fee = (amountDiffFromLow * lowHighFeeDiff) / lowHighAmountDiff)
        const temp1 = biggystring.bns.mul(amountDiffFromLow, lowHighFeeDiff);
        const addFeeToLow = biggystring.bns.div(temp1, lowHighAmountDiff);
        gasPrice = biggystring.bns.add(gasPriceObj.standardFeeLow, addFeeToLow);
        break;
      case ES_FEE_HIGH:
        gasPrice = networkFeeForGasPrice.gasPrice.highFee;
        break;
      default:
        throw new Error(`Invalid networkFeeOption`);
    }
    const out = { gasLimit, gasPrice };
    return out;
  } else {
    throw new Error('ErrorInvalidSpendInfo');
  }
}

/**
 * Created by paul on 8/27/17.
 */



const NetworkFeesSchema = {
  type: 'object',
  additionalProperties: {
    type: 'object',
    properties: {
      gasLimit: {
        type: 'object',
        properties: {
          regularTransaction: { type: 'string' },
          tokenTransaction: { type: 'string' }
        },
        required: ['regularTransaction', 'tokenTransaction']
      },
      gasPrice: {
        type: 'object',
        properties: {
          lowFee: { type: 'string' },
          standardFeeLow: { type: 'string' },
          standardFeeHigh: { type: 'string' },
          standardFeeLowAmount: { type: 'string' },
          standardFeeHighAmount: { type: 'string' },
          highFee: { type: 'string' }
        },
        required: ['lowFee', 'standardFeeLow', 'standardFeeHigh', 'standardFeeLowAmount', 'standardFeeHighAmount', 'highFee']
      }
    },
    required: ['gasLimit']
  }
};

const EthGasStationSchema = {
  type: 'object',
  properties: {
    safeLow: { type: 'number' },
    average: { type: 'number' },
    fastest: { type: 'number' }
  },
  required: ['safeLow', 'average', 'fastest']
};

const CustomTokenSchema = {
  type: 'object',
  properties: {
    currencyCode: { type: 'string' },
    currencyName: { type: 'string' },
    multiplier: { type: 'string' },
    contractAddress: { type: 'string' }
  },
  required: ['currencyCode', 'currencyName', 'multiplier', 'contractAddress']
};

const DATA_STORE_FOLDER = 'txEngineFolder';
const DATA_STORE_FILE = 'walletLocalData.json';
const PRIMARY_CURRENCY$1 = currencyInfo.currencyCode;

const defaultNetworkFees = {
  default: {
    gasLimit: {
      regularTransaction: '21001',
      tokenTransaction: '37123'
    },
    gasPrice: {
      lowFee: '1000000001',
      standardFeeLow: '40000000001',
      standardFeeHigh: '300000000001',
      standardFeeLowAmount: '100000000000000000',
      standardFeeHighAmount: '10000000000000000000',
      highFee: '40000000001'
    }
  },
  '1983987abc9837fbabc0982347ad828': {
    gasLimit: {
      regularTransaction: '21002',
      tokenTransaction: '37124'
    },
    gasPrice: {
      lowFee: '1000000002',
      standardFeeLow: '40000000002',
      standardFeeHigh: '300000000002',
      standardFeeLowAmount: '200000000000000000',
      standardFeeHighAmount: '20000000000000000000',
      highFee: '40000000002'
    }
  },
  '2983987abc9837fbabc0982347ad828': {
    gasLimit: {
      regularTransaction: '21002',
      tokenTransaction: '37124'
    }
  }
};

class WalletLocalData {

  constructor(jsonString) {
    this.blockHeight = 0;

    const totalBalances = {};
    this.totalBalances = totalBalances;

    this.nextNonce = '0';

    this.lastAddressQueryHeight = 0;

    // Dumb extra local var needed to make Flow happy
    const transactionsObj = {};
    this.transactionsObj = transactionsObj;

    this.networkFees = defaultNetworkFees;

    this.ethereumAddress = '';
    this.enabledTokens = [PRIMARY_CURRENCY$1];
    if (jsonString !== null) {
      const data = JSON.parse(jsonString);

      if (typeof data.blockHeight === 'number') {
        this.blockHeight = data.blockHeight;
      }
      if (typeof data.lastAddressQueryHeight === 'string') {
        this.lastAddressQueryHeight = data.lastAddressQueryHeight;
      }
      if (typeof data.nextNonce === 'string') this.nextNonce = data.nextNonce;
      if (typeof data.ethereumAddress === 'string') {
        this.ethereumAddress = data.ethereumAddress;
      }
      if (typeof data.totalBalances !== 'undefined') {
        this.totalBalances = data.totalBalances;
      }
      if (typeof data.enabledTokens !== 'undefined') {
        this.enabledTokens = data.enabledTokens;
      }
      if (typeof data.networkFees !== 'undefined') {
        this.networkFees = data.networkFees;
      }
      if (typeof data.transactionsObj !== 'undefined') {
        this.transactionsObj = data.transactionsObj;
      }
    }
  }
}

/**
 * Created by paul on 7/7/17.
 */
const Buffer$1 = require('buffer/').Buffer;
const abi = require('../lib/export-fixes-bundle.js').ABI;
const ethWallet$1 = require('../lib/export-fixes-bundle.js').Wallet;
const EthereumTx = require('../lib/export-fixes-bundle.js').Transaction;

const ADDRESS_POLL_MILLISECONDS = 3000;
const BLOCKHEIGHT_POLL_MILLISECONDS = 5000;
const NETWORKFEES_POLL_MILLISECONDS = 60 * 10 * 1000; // 10 minutes
const SAVE_DATASTORE_MILLISECONDS = 10000;
// const ADDRESS_QUERY_LOOKBACK_BLOCKS = '8' // ~ 2 minutes
const ADDRESS_QUERY_LOOKBACK_BLOCKS = 4 * 60 * 24 * 7; // ~ one week

const PRIMARY_CURRENCY = currencyInfo.currencyCode;
const CHECK_UNCONFIRMED = true;
const INFO_SERVERS = ['https://info1.edgesecure.co:8444'];

function unpadAddress(address) {
  const normalizedAddress = normalizeAddress(address);
  const unpadded = normalizedAddress.slice(24);
  const out = '0x' + unpadded;
  return out;
}

function padAddress(address) {
  const normalizedAddress = normalizeAddress(address);
  const padding = 64 - normalizedAddress.length;
  const zeroString = '0000000000000000000000000000000000000000000000000000000000000000';
  const out = '0x' + zeroString.slice(0, padding) + normalizedAddress;
  return out;
}

class EthereumParams {

  constructor(from, to, gas, gasPrice, gasUsed, cumulativeGasUsed, errorVal, tokenRecipientAddress) {
    this.from = from;
    this.to = to;
    this.gas = gas;
    this.gasPrice = gasPrice;
    this.gasUsed = gasUsed;
    this.errorVal = errorVal;
    this.cumulativeGasUsed = cumulativeGasUsed;
    if (typeof tokenRecipientAddress === 'string') {
      this.tokenRecipientAddress = tokenRecipientAddress;
    } else {
      this.tokenRecipientAddress = null;
    }
  }
}

class EthereumEngine {
  // Each currency code can be a 0-1 value
  constructor(io_, walletInfo, opts) {
    // Validate that we are a valid EdgeCurrencyEngine:
    const test = this;

    const { walletLocalFolder, callbacks } = opts;

    this.io = io_;
    this.engineOn = false;
    this.addressesChecked = false;
    this.tokenCheckStatus = {};
    this.walletLocalDataDirty = false;
    this.transactionsChangedArray = [];
    this.walletInfo = walletInfo;
    this.walletId = walletInfo.id ? `${walletInfo.id} - ` : '';
    this.currencyInfo = currencyInfo;
    this.allTokens = currencyInfo.metaTokens.slice(0);
    this.customTokens = [];
    this.timers = {};

    if (typeof opts.optionalSettings !== 'undefined') {
      this.currentSettings = opts.optionalSettings;
    } else {
      this.currentSettings = this.currencyInfo.defaultSettings;
    }

    // Hard coded for testing
    // this.walletInfo.keys.ethereumKey = '389b07b3466eed587d6bdae09a3613611de9add2635432d6cd1521af7bbc3757'
    // this.walletInfo.keys.ethereumAddress = '0x9fa817e5A48DD1adcA7BEc59aa6E3B1F5C4BeA9a'
    this.edgeTxLibCallbacks = callbacks;
    this.walletLocalFolder = walletLocalFolder;

    // Fix messed-up wallets that have a private key in the wrong place:
    if (typeof this.walletInfo.keys.ethereumKey !== 'string') {
      if (walletInfo.keys.keys && walletInfo.keys.keys.ethereumKey) {
        this.walletInfo.keys.ethereumKey = walletInfo.keys.keys.ethereumKey;
      }
    }

    // Fix messed-up wallets that have a public address in the wrong place:
    if (typeof this.walletInfo.keys.ethereumAddress !== 'string') {
      if (walletInfo.keys.ethereumPublicAddress) {
        this.walletInfo.keys.ethereumAddress = walletInfo.keys.ethereumPublicAddress;
      } else if (walletInfo.keys.keys && walletInfo.keys.keys.ethereumPublicAddress) {
        this.walletInfo.keys.ethereumAddress = walletInfo.keys.keys.ethereumPublicAddress;
      } else {
        const privKey = Buffer$1.from(this.walletInfo.keys.ethereumKey, 'hex');
        const wallet = ethWallet$1.fromPrivateKey(privKey);
        this.walletInfo.keys.ethereumAddress = wallet.getAddressString();
      }
    }
    this.log(`Created Wallet Type ${this.walletInfo.type} for Currency Plugin ${this.currencyInfo.pluginName} `);
  }

  // *************************************
  // Private methods
  // *************************************
  async fetchGetEtherscan(cmd) {
    let apiKey = '';
    if (global.etherscanApiKey && global.etherscanApiKey.length > 5) {
      apiKey = '&apikey=' + global.etherscanApiKey;
    }
    const url = sprintfJs.sprintf('%s/api%s%s', this.currentSettings.otherSettings.etherscanApiServers[0], cmd, apiKey);
    return this.fetchGet(url);
  }

  async fetchGet(url) {
    const response = await this.io.fetch(url, {
      method: 'GET'
    });
    if (!response.ok) {
      const cleanUrl = url.replace(global.etherscanApiKey, 'private');
      throw new Error(`The server returned error code ${response.status} for ${cleanUrl}`);
    }
    return response.json();
  }

  async fetchPostBlockcypher(cmd, body) {
    let apiKey = '';
    if (global.blockcypherApiKey && global.blockcypherApiKey.length > 5) {
      apiKey = '&token=' + global.blockcypherApiKey;
    }
    const url = sprintfJs.sprintf('%s/%s%s', this.currentSettings.otherSettings.blockcypherApiServers[0], cmd, apiKey);
    const response = await this.io.fetch(url, {
      headers: {
        Accept: 'application/json',
        'Content-Type': 'application/json'
      },
      method: 'POST',
      body: JSON.stringify(body)
    });
    return response.json();
  }

  // *************************************
  // Poll on the blockheight
  // *************************************
  async blockHeightInnerLoop() {
    try {
      const jsonObj = await this.fetchGetEtherscan('?module=proxy&action=eth_blockNumber');
      const valid = validateObject(jsonObj, {
        type: 'object',
        properties: {
          result: { type: 'string' }
        },
        required: ['result']
      });
      if (valid) {
        const blockHeight = parseInt(jsonObj.result, 16);
        this.log(`Got block height ${blockHeight}`);
        if (this.walletLocalData.blockHeight !== blockHeight) {
          this.walletLocalData.blockHeight = blockHeight; // Convert to decimal
          this.walletLocalDataDirty = true;
          this.edgeTxLibCallbacks.onBlockHeightChanged(this.walletLocalData.blockHeight);
        }
      }
    } catch (err) {
      this.log('Error fetching height: ' + err);
    }
  }

  processEtherscanTransaction(tx) {
    let netNativeAmount; // Amount received into wallet
    const ourReceiveAddresses = [];

    const nativeNetworkFee = biggystring.bns.mul(tx.gasPrice, tx.gasUsed);

    if (tx.from.toLowerCase() === this.walletLocalData.ethereumAddress.toLowerCase()) {
      netNativeAmount = biggystring.bns.sub('0', tx.value);

      // For spends, include the network fee in the transaction amount
      netNativeAmount = biggystring.bns.sub(netNativeAmount, nativeNetworkFee);

      if (biggystring.bns.gte(tx.nonce, this.walletLocalData.nextNonce)) {
        this.walletLocalData.nextNonce = biggystring.bns.add(tx.nonce, '1');
      }
    } else {
      netNativeAmount = biggystring.bns.add('0', tx.value);
      ourReceiveAddresses.push(this.walletLocalData.ethereumAddress.toLowerCase());
    }

    const ethParams = new EthereumParams([tx.from], [tx.to], tx.gas, tx.gasPrice, tx.gasUsed, tx.cumulativeGasUsed, parseInt(tx.isError), null);

    const edgeTransaction = {
      txid: tx.hash,
      date: parseInt(tx.timeStamp),
      currencyCode: 'ETH',
      blockHeight: parseInt(tx.blockNumber),
      nativeAmount: netNativeAmount,
      networkFee: nativeNetworkFee,
      ourReceiveAddresses,
      signedTx: 'unsigned_right_now',
      otherParams: ethParams
    };

    const idx = this.findTransaction(PRIMARY_CURRENCY, tx.hash);
    if (idx === -1) {
      this.log(sprintfJs.sprintf('New transaction: %s', tx.hash));

      // New transaction not in database
      this.addTransaction(PRIMARY_CURRENCY, edgeTransaction);

      this.edgeTxLibCallbacks.onTransactionsChanged(this.transactionsChangedArray);
      this.transactionsChangedArray = [];
    } else {
      // Already have this tx in the database. See if anything changed
      const txs = this.walletLocalData.transactionsObj;
      const transactionsArray = txs[PRIMARY_CURRENCY];
      const edgeTx = transactionsArray[idx];

      if (edgeTx.blockHeight !== edgeTransaction.blockHeight || edgeTx.networkFee !== edgeTransaction.networkFee || edgeTx.nativeAmount !== edgeTransaction.nativeAmount || edgeTx.otherParams.errorVal !== edgeTransaction.otherParams.errorVal) {
        this.log(sprintfJs.sprintf('Update transaction: %s height:%s', tx.hash, tx.blockNumber));
        this.updateTransaction(PRIMARY_CURRENCY, edgeTransaction, idx);
        this.edgeTxLibCallbacks.onTransactionsChanged(this.transactionsChangedArray);
        this.transactionsChangedArray = [];
      } else {
        // this.log(sprintf('Old transaction. No Update: %s', tx.hash))
      }
    }
  }

  processEtherscanTokenTransaction(tx, currencyCode) {
    let netNativeAmount; // Amount received into wallet
    const ourReceiveAddresses = [];

    // const nativeValueBN = new BN(tx.value, 10)
    const paddedAddress = padAddress(this.walletLocalData.ethereumAddress);
    let fromAddress;
    let toAddress;

    if (tx.topics[1] === paddedAddress) {
      netNativeAmount = biggystring.bns.sub('0', tx.data);
      fromAddress = this.walletLocalData.ethereumAddress;
      toAddress = unpadAddress(tx.topics[2]);
    } else {
      fromAddress = unpadAddress(tx.topics[1]);
      toAddress = this.walletLocalData.ethereumAddress;
      netNativeAmount = biggystring.bns.add('0', tx.data);
      ourReceiveAddresses.push(this.walletLocalData.ethereumAddress.toLowerCase());
    }

    if (netNativeAmount.length > 50) {
      // Etherscan occasionally send back a transactino with a corrupt amount in tx.data. Ignore this tx.
      return;
    }

    const ethParams = new EthereumParams([fromAddress], [toAddress], '', tx.gasPrice, tx.gasUsed, '', 0, null);

    const edgeTransaction = {
      txid: tx.transactionHash,
      date: parseInt(tx.timeStamp),
      currencyCode,
      blockHeight: parseInt(biggystring.bns.add('0', tx.blockNumber)),
      nativeAmount: netNativeAmount,
      networkFee: '0',
      ourReceiveAddresses,
      signedTx: 'unsigned_right_now',
      otherParams: ethParams
    };

    const idx = this.findTransaction(currencyCode, tx.transactionHash);
    if (idx === -1) {
      this.log(sprintfJs.sprintf('New token transaction: %s', tx.transactionHash));

      // New transaction not in database
      this.addTransaction(currencyCode, edgeTransaction);

      this.edgeTxLibCallbacks.onTransactionsChanged(this.transactionsChangedArray);
      this.transactionsChangedArray = [];
    } else {
      // Already have this tx in the database. See if anything changed
      const txs = this.walletLocalData.transactionsObj;
      const transactionsArray = txs[currencyCode];
      const edgeTx = transactionsArray[idx];

      if (edgeTx.blockHeight !== edgeTransaction.blockHeight || edgeTx.networkFee !== edgeTransaction.networkFee || edgeTx.nativeAmount !== edgeTransaction.nativeAmount || edgeTx.otherParams.errorVal !== edgeTransaction.otherParams.errorVal) {
        this.log(sprintfJs.sprintf('Update token transaction: %s height:%s', edgeTx.txid, edgeTx.blockHeight));
        this.updateTransaction(currencyCode, edgeTransaction, idx);
        this.edgeTxLibCallbacks.onTransactionsChanged(this.transactionsChangedArray);
        this.transactionsChangedArray = [];
      } else {
        // this.log(sprintf('Old transaction. No Update: %s', edgeTx.txid))
      }
    }
  }

  processUnconfirmedTransaction(tx) {
    const fromAddress = '0x' + tx.inputs[0].addresses[0];
    const toAddress = '0x' + tx.outputs[0].addresses[0];
    const epochTime = Date.parse(tx.received) / 1000;
    const ourReceiveAddresses = [];

    let nativeAmount;
    if (normalizeAddress(fromAddress) === normalizeAddress(this.walletLocalData.ethereumAddress)) {
      nativeAmount = (0 - tx.total).toString(10);
      nativeAmount = biggystring.bns.sub(nativeAmount, tx.fees.toString(10));
    } else {
      nativeAmount = tx.total.toString(10);
      ourReceiveAddresses.push(this.walletLocalData.ethereumAddress);
    }

    const ethParams = new EthereumParams([fromAddress], [toAddress], '', '', tx.fees.toString(10), '', 0, null);

    const edgeTransaction = {
      txid: addHexPrefix(tx.hash),
      date: epochTime,
      currencyCode: 'ETH',
      blockHeight: tx.block_height,
      nativeAmount,
      networkFee: tx.fees.toString(10),
      ourReceiveAddresses,
      signedTx: 'iwassignedyoucantrustme',
      otherParams: ethParams
    };

    const idx = this.findTransaction(PRIMARY_CURRENCY, tx.hash);
    if (idx === -1) {
      this.log(sprintfJs.sprintf('processUnconfirmedTransaction: New transaction: %s', tx.hash));

      // New transaction not in database
      this.addTransaction(PRIMARY_CURRENCY, edgeTransaction);

      this.edgeTxLibCallbacks.onTransactionsChanged(this.transactionsChangedArray);
      this.transactionsChangedArray = [];
    } else {
      // Already have this tx in the database. See if anything changed
      // const transactionsArray:Array<EdgeTransaction> = this.walletLocalData.transactionsObj[ PRIMARY_CURRENCY ]
      // const edgeTx:EdgeTransaction = transactionsArray[ idx ]
      //
      // if (edgeTx.blockHeight < tx.block_height || edgeTx.date > epochTime) {
      //   this.log(sprintf('processUnconfirmedTransaction: Update transaction: %s height:%s', tx.hash, tx.blockNumber))
      //   this.updateTransaction(PRIMARY_CURRENCY, edgeTransaction, idx)
      //   this.edgeTxLibCallbacks.onTransactionsChanged(
      //     this.transactionsChangedArray
      //   )
      //   this.transactionsChangedArray = []
      // } else {
      // this.log(sprintf('processUnconfirmedTransaction: Old transaction. No Update: %s', tx.hash))
      // }
    }
  }

  async checkAddressFetch(tk, url) {
    let checkAddressSuccess = true;
    let jsonObj = {};
    let valid = false;

    try {
      jsonObj = await this.fetchGetEtherscan(url);
      valid = validateObject(jsonObj, {
        type: 'object',
        properties: {
          result: { type: 'string' }
        },
        required: ['result']
      });
      if (valid) {
        const balance = jsonObj.result;

        if (typeof this.walletLocalData.totalBalances[tk] === 'undefined') {
          this.walletLocalData.totalBalances[tk] = '0';
        }
        if (!biggystring.bns.eq(balance, this.walletLocalData.totalBalances[tk])) {
          this.walletLocalData.totalBalances[tk] = balance;
          this.log(tk + ': token Address balance: ' + balance);
          this.edgeTxLibCallbacks.onBalanceChanged(tk, balance);
        }
      } else {
        checkAddressSuccess = false;
      }
    } catch (e) {
      checkAddressSuccess = false;
    }
    return checkAddressSuccess;
  }

  async checkTransactionsFetch() {
    const address = this.walletLocalData.ethereumAddress;
    const endBlock = 999999999;
    let startBlock = 0;
    let checkAddressSuccess = true;
    let url = '';
    let jsonObj = {};
    let valid = false;
    if (this.walletLocalData.lastAddressQueryHeight > ADDRESS_QUERY_LOOKBACK_BLOCKS) {
      // Only query for transactions as far back as ADDRESS_QUERY_LOOKBACK_BLOCKS from the last time we queried transactions
      startBlock = this.walletLocalData.lastAddressQueryHeight - ADDRESS_QUERY_LOOKBACK_BLOCKS;
    }

    try {
      url = sprintfJs.sprintf('?module=account&action=txlist&address=%s&startblock=%d&endblock=%d&sort=asc', address, startBlock, endBlock);
      jsonObj = await this.fetchGetEtherscan(url);
      valid = validateObject(jsonObj, {
        type: 'object',
        properties: {
          result: {
            type: 'array',
            items: {
              type: 'object',
              properties: {
                blockNumber: { type: 'string' },
                timeStamp: { type: 'string' },
                hash: { type: 'string' },
                from: { type: 'string' },
                to: { type: 'string' },
                nonce: { type: 'string' },
                value: { type: 'string' },
                gas: { type: 'string' },
                gasPrice: { type: 'string' },
                cumulativeGasUsed: { type: 'string' },
                gasUsed: { type: 'string' },
                confirmations: { type: 'string' }
              },
              required: ['blockNumber', 'timeStamp', 'hash', 'from', 'to', 'nonce', 'value', 'gas', 'gasPrice', 'cumulativeGasUsed', 'gasUsed', 'confirmations']
            }
          }
        },
        required: ['result']
      });

      if (valid) {
        const transactions = jsonObj.result;
        this.log('Fetched transactions count: ' + transactions.length);

        // Get transactions
        // Iterate over transactions in address
        for (let i = 0; i < transactions.length; i++) {
          const tx = transactions[i];
          this.processEtherscanTransaction(tx);
          this.tokenCheckStatus[PRIMARY_CURRENCY] = (i + 1) / transactions.length;
          if (i % 10 === 0) {
            this.updateOnAddressesChecked();
          }
        }
        if (transactions.length === 0) {
          this.tokenCheckStatus[PRIMARY_CURRENCY] = 1;
        }
        this.updateOnAddressesChecked();
      } else {
        checkAddressSuccess = false;
      }
    } catch (e) {
      this.log(e);
      checkAddressSuccess = false;
    }
    return checkAddressSuccess;
  }

  updateOnAddressesChecked() {
    if (this.addressesChecked) {
      return;
    }
    const activeTokens = [];

    for (const tokenCode of this.walletLocalData.enabledTokens) {
      const ti = this.getTokenInfo(tokenCode);
      if (ti) {
        activeTokens.push(tokenCode);
      }
    }

    const perTokenSlice = 1 / activeTokens.length;
    let numCompleteStatus = 0;
    let totalStatus = 0;
    for (const token of activeTokens) {
      const status = this.tokenCheckStatus[token];
      totalStatus += status * perTokenSlice;
      if (status === 1) {
        numCompleteStatus++;
      }
    }
    if (numCompleteStatus === activeTokens.length) {
      this.addressesChecked = true;
      this.edgeTxLibCallbacks.onAddressesChecked(1);
      this.walletLocalData.lastAddressQueryHeight = this.walletLocalData.blockHeight;
    } else {
      this.edgeTxLibCallbacks.onAddressesChecked(totalStatus);
    }
  }

  async checkTokenTransactionsFetch(currencyCode) {
    const address = padAddress(this.walletLocalData.ethereumAddress);
    let startBlock = 0;
    let checkAddressSuccess = true;
    let url = '';
    let jsonObj = {};
    let valid = false;
    if (this.walletLocalData.lastAddressQueryHeight > ADDRESS_QUERY_LOOKBACK_BLOCKS) {
      // Only query for transactions as far back as ADDRESS_QUERY_LOOKBACK_BLOCKS from the last time we queried transactions
      startBlock = this.walletLocalData.lastAddressQueryHeight - ADDRESS_QUERY_LOOKBACK_BLOCKS;
    }

    const tokenInfo = this.getTokenInfo(currencyCode);
    let contractAddress = '';
    if (tokenInfo && typeof tokenInfo.contractAddress === 'string') {
      contractAddress = tokenInfo.contractAddress;
    } else {
      return;
    }

    try {
      url = sprintfJs.sprintf('?module=logs&action=getLogs&fromBlock=%d&toBlock=latest&address=%s&topic0=0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef&topic0_1_opr=and&topic1=%s&topic1_2_opr=or&topic2=%s', startBlock, contractAddress, address, address);
      jsonObj = await this.fetchGetEtherscan(url);
      valid = validateObject(jsonObj, {
        type: 'object',
        properties: {
          result: {
            type: 'array',
            items: {
              type: 'object',
              properties: {
                data: { type: 'string' },
                blockNumber: { type: 'string' },
                timeStamp: { type: 'string' },
                transactionHash: { type: 'string' },
                gasPrice: { type: 'string' },
                gasUsed: { type: 'string' },
                topics: {
                  type: 'array',
                  items: { type: 'string' }
                }
              },
              required: ['data', 'blockNumber', 'timeStamp', 'transactionHash', 'gasPrice', 'gasUsed', 'topics']
            }
          }
        },
        required: ['result']
      });

      if (valid) {
        const transactions = jsonObj.result;
        this.log(`Fetched token ${tokenInfo.currencyCode} transactions count: ${transactions.length}`);

        // Get transactions
        // Iterate over transactions in address
        for (let i = 0; i < transactions.length; i++) {
          const tx = transactions[i];
          this.processEtherscanTokenTransaction(tx, currencyCode);
          this.tokenCheckStatus[currencyCode] = (i + 1) / transactions.length;
          if (i % 10 === 0) {
            this.updateOnAddressesChecked();
          }
        }
        if (transactions.length === 0) {
          this.tokenCheckStatus[currencyCode] = 1;
        }
        this.updateOnAddressesChecked();
      } else {
        checkAddressSuccess = false;
      }
    } catch (e) {
      this.log(e);
      checkAddressSuccess = false;
    }
    return checkAddressSuccess;
  }

  async checkUnconfirmedTransactionsFetch() {
    const address = normalizeAddress(this.walletLocalData.ethereumAddress);
    const url = sprintfJs.sprintf('%s/v1/eth/main/txs/%s', this.currentSettings.otherSettings.superethServers[0], address);
    let jsonObj = null;
    try {
      jsonObj = await this.fetchGet(url);
    } catch (e) {
      this.log(e);
      this.log('Failed to fetch unconfirmed transactions');
      return;
    }

    const valid = validateObject(jsonObj, {
      type: 'array',
      items: {
        type: 'object',
        properties: {
          block_height: { type: 'number' },
          fees: { type: 'number' },
          received: { type: 'string' },
          addresses: {
            type: 'array',
            items: { type: 'string' }
          },
          inputs: {
            type: 'array',
            items: {
              type: 'object',
              properties: {
                addresses: {
                  type: 'array',
                  items: { type: 'string' }
                }
              },
              required: ['addresses']
            }
          },
          outputs: {
            type: 'array',
            items: {
              type: 'object',
              properties: {
                addresses: {
                  type: 'array',
                  items: { type: 'string' }
                }
              },
              required: ['addresses']
            }
          }
        },
        required: ['fees', 'received', 'addresses', 'inputs', 'outputs']
      }
    });

    if (valid) {
      const transactions = jsonObj;

      for (const tx of transactions) {
        if (normalizeAddress(tx.inputs[0].addresses[0]) === address || normalizeAddress(tx.outputs[0].addresses[0]) === address) {
          this.processUnconfirmedTransaction(tx);
        }
      }
    } else {
      this.log('Invalid data for unconfirmed transactions');
    }
  }

  // **********************************************
  // Check all addresses for new transactions
  // **********************************************
  async checkAddressesInnerLoop() {
    const address = this.walletLocalData.ethereumAddress;
    try {
      // Ethereum only has one address
      let url = '';
      const promiseArray = [];

      // ************************************
      // Fetch token balances
      // ************************************
      // https://api.etherscan.io/api?module=account&action=tokenbalance&contractaddress=0x57d90b64a1a57749b0f932f1a3395792e12e7055&address=0xe04f27eb70e025b78871a2ad7eabe85e61212761&tag=latest&apikey=YourApiKeyToken
      for (const tk of this.walletLocalData.enabledTokens) {
        if (tk === PRIMARY_CURRENCY) {
          url = sprintfJs.sprintf('?module=account&action=balance&address=%s&tag=latest', address);
        } else {
          if (this.getTokenStatus(tk)) {
            const tokenInfo = this.getTokenInfo(tk);
            if (tokenInfo && typeof tokenInfo.contractAddress === 'string') {
              url = sprintfJs.sprintf('?module=account&action=tokenbalance&contractaddress=%s&address=%s&tag=latest', tokenInfo.contractAddress, this.walletLocalData.ethereumAddress);
              promiseArray.push(this.checkTokenTransactionsFetch(tk));
            } else {
              continue;
            }
          } else {
            continue;
          }
        }
        promiseArray.push(this.checkAddressFetch(tk, url));
      }

      promiseArray.push(this.checkTransactionsFetch());
      if (CHECK_UNCONFIRMED) {
        promiseArray.push(this.checkUnconfirmedTransactionsFetch());
      }
      await Promise.all(promiseArray);
    } catch (e) {
      this.log('Error fetching address transactions: ' + address);
    }
  }

  findTransaction(currencyCode, txid) {
    if (typeof this.walletLocalData.transactionsObj[currencyCode] === 'undefined') {
      return -1;
    }

    const currency = this.walletLocalData.transactionsObj[currencyCode];
    return currency.findIndex(element => {
      return normalizeAddress(element.txid) === normalizeAddress(txid);
    });
  }

  sortTxByDate(a, b) {
    return b.date - a.date;
  }

  addTransaction(currencyCode, edgeTransaction) {
    // Add or update tx in transactionsObj
    const idx = this.findTransaction(currencyCode, edgeTransaction.txid);

    if (idx === -1) {
      this.log('addTransaction: adding and sorting:' + edgeTransaction.txid);
      if (typeof this.walletLocalData.transactionsObj[currencyCode] === 'undefined') {
        this.walletLocalData.transactionsObj[currencyCode] = [];
      }
      this.walletLocalData.transactionsObj[currencyCode].push(edgeTransaction);

      // Sort
      this.walletLocalData.transactionsObj[currencyCode].sort(this.sortTxByDate);
      this.walletLocalDataDirty = true;
      this.transactionsChangedArray.push(edgeTransaction);
    } else {
      this.updateTransaction(currencyCode, edgeTransaction, idx);
    }
  }

  updateTransaction(currencyCode, edgeTransaction, idx) {
    // Update the transaction
    this.walletLocalData.transactionsObj[currencyCode][idx] = edgeTransaction;
    this.walletLocalDataDirty = true;
    this.transactionsChangedArray.push(edgeTransaction);
    this.log('updateTransaction:' + edgeTransaction.txid);
  }

  // *************************************
  // Save the wallet data store
  // *************************************
  async saveWalletLoop() {
    if (this.walletLocalDataDirty) {
      try {
        this.log('walletLocalDataDirty. Saving...');
        const walletJson = JSON.stringify(this.walletLocalData);
        await this.walletLocalFolder.folder(DATA_STORE_FOLDER).file(DATA_STORE_FILE).setText(walletJson);
        this.walletLocalDataDirty = false;
      } catch (err) {
        this.log(err);
      }
    }
  }

  async checkUpdateNetworkFees() {
    try {
      const url = sprintfJs.sprintf('%s/v1/networkFees/ETH', INFO_SERVERS[0]);
      const jsonObj = await this.fetchGet(url);
      const valid = validateObject(jsonObj, NetworkFeesSchema);

      if (valid) {
        this.walletLocalData.networkFees = jsonObj;
      } else {
        this.log('Error: Fetched invalid networkFees');
      }
    } catch (err) {
      this.log('Error fetching networkFees:');
      this.log(err);
    }

    try {
      const url = sprintfJs.sprintf('https://www.ethgasstation.info/json/ethgasAPI.json');
      const jsonObj = await this.fetchGet(url);
      const valid = validateObject(jsonObj, EthGasStationSchema);

      if (valid) {
        const fees = this.walletLocalData.networkFees;
        const ethereumFee = fees['default'];
        if (!ethereumFee.gasPrice) {
          return;
        }
        const gasPrice = ethereumFee.gasPrice;

        const safeLow = Math.floor(jsonObj.safeLow / 10);
        let average = Math.floor(jsonObj.average / 10);
        let fastest = Math.floor(jsonObj.fastest / 10);

        // Sanity checks
        if (safeLow < 1 || safeLow > 300) {
          console.log('Invalid safeLow value from EthGasStation');
          return;
        }
        if (average < 1 || average > 300) {
          console.log('Invalid average value from EthGasStation');
          return;
        }
        if (fastest < 1 || fastest > 300) {
          console.log('Invalid fastest value from EthGasStation');
          return;
        }

        gasPrice.lowFee = (safeLow * 1000000000).toString();

        if (average <= safeLow) average = safeLow + 1;
        gasPrice.standardFeeLow = (average * 1000000000).toString();

        if (fastest <= average) fastest = average + 1;
        gasPrice.highFee = (fastest * 1000000000).toString();

        // We use a value that is somewhere in between average and fastest for the standardFeeHigh
        gasPrice.standardFeeHigh = (Math.floor((average + fastest) * 0.75) * 1000000000).toString();
      } else {
        this.log('Error: Fetched invalid networkFees from EthGasStation');
      }
    } catch (err) {
      this.log('Error fetching networkFees from EthGasStation');
      this.log(err);
    }
  }

  doInitialCallbacks() {
    for (const currencyCode of this.walletLocalData.enabledTokens) {
      try {
        this.edgeTxLibCallbacks.onTransactionsChanged(this.walletLocalData.transactionsObj[currencyCode]);
        this.edgeTxLibCallbacks.onBalanceChanged(currencyCode, this.walletLocalData.totalBalances[currencyCode]);
      } catch (e) {
        this.log('Error for currencyCode', currencyCode, e);
      }
    }
  }

  getTokenInfo(token) {
    return this.allTokens.find(element => {
      return element.currencyCode === token;
    });
  }

  async addToLoop(func, timer) {
    try {
      // $FlowFixMe
      await this[func]();
    } catch (e) {
      this.log('Error in Loop:', func, e);
    }
    if (this.engineOn) {
      this.timers[func] = setTimeout(() => {
        if (this.engineOn) {
          this.addToLoop(func, timer);
        }
      }, timer);
    }
    return true;
  }

  log(...text) {
    text[0] = `${this.walletId}${text[0]}`;
    console.log(...text);
  }

  // *************************************
  // Public methods
  // *************************************

  updateSettings(settings) {
    this.currentSettings = settings;
  }

  async startEngine() {
    this.engineOn = true;
    this.doInitialCallbacks();
    this.addToLoop('blockHeightInnerLoop', BLOCKHEIGHT_POLL_MILLISECONDS);
    this.addToLoop('checkAddressesInnerLoop', ADDRESS_POLL_MILLISECONDS);
    this.addToLoop('saveWalletLoop', SAVE_DATASTORE_MILLISECONDS);
    this.addToLoop('checkUpdateNetworkFees', NETWORKFEES_POLL_MILLISECONDS);
  }

  async killEngine() {
    // Set status flag to false
    this.engineOn = false;
    // Clear Inner loops timers
    for (const timer in this.timers) {
      clearTimeout(this.timers[timer]);
    }
    this.timers = {};
  }

  async resyncBlockchain() {
    await this.killEngine();
    const temp = JSON.stringify({
      enabledTokens: this.walletLocalData.enabledTokens,
      networkFees: this.walletLocalData.networkFees,
      ethereumAddress: this.walletLocalData.ethereumAddress
    });
    this.walletLocalData = new WalletLocalData(temp);
    this.walletLocalDataDirty = true;
    await this.saveWalletLoop();
    await this.startEngine();
  }

  // synchronous
  getBlockHeight() {
    return parseInt(this.walletLocalData.blockHeight);
  }

  enableTokensSync(tokens) {
    for (const token of tokens) {
      if (this.walletLocalData.enabledTokens.indexOf(token) === -1) {
        this.walletLocalData.enabledTokens.push(token);
      }
    }
  }

  // asynchronous
  async enableTokens(tokens) {
    this.enableTokensSync(tokens);
  }

  disableTokensSync(tokens) {
    for (const token of tokens) {
      const index = this.walletLocalData.enabledTokens.indexOf(token);
      if (index !== -1) {
        this.walletLocalData.enabledTokens.splice(index, 1);
      }
    }
  }

  // asynchronous
  async disableTokens(tokens) {
    this.disableTokensSync(tokens);
  }

  async getEnabledTokens() {
    return this.walletLocalData.enabledTokens;
  }

  async addCustomToken(tokenObj) {
    const valid = validateObject(tokenObj, CustomTokenSchema);

    if (valid) {
      const ethTokenObj = tokenObj;
      // If token is already in currencyInfo, error as it cannot be changed
      for (const tk of this.currencyInfo.metaTokens) {
        if (tk.currencyCode.toLowerCase() === ethTokenObj.currencyCode.toLowerCase() || tk.currencyName.toLowerCase() === ethTokenObj.currencyName.toLowerCase()) {
          throw new Error('ErrorCannotModifyToken');
        }
      }

      // Validate the token object
      if (ethTokenObj.currencyCode.toUpperCase() !== ethTokenObj.currencyCode) {
        throw new Error('ErrorInvalidCurrencyCode');
      }
      if (ethTokenObj.currencyCode.length < 2 || ethTokenObj.currencyCode.length > 7) {
        throw new Error('ErrorInvalidCurrencyCodeLength');
      }
      if (ethTokenObj.currencyName.length < 3 || ethTokenObj.currencyName.length > 20) {
        throw new Error('ErrorInvalidCurrencyNameLength');
      }
      if (biggystring.bns.lt(ethTokenObj.multiplier, '1') || biggystring.bns.gt(ethTokenObj.multiplier, '100000000000000000000000000000000')) {
        throw new Error('ErrorInvalidMultiplier');
      }
      let contractAddress = ethTokenObj.contractAddress.replace('0x', '').toLowerCase();
      if (!isHex(contractAddress) || contractAddress.length !== 40) {
        throw new Error('ErrorInvalidContractAddress');
      }
      contractAddress = '0x' + contractAddress;

      for (const tk of this.customTokens) {
        if (tk.currencyCode.toLowerCase() === ethTokenObj.currencyCode.toLowerCase() || tk.currencyName.toLowerCase() === ethTokenObj.currencyName.toLowerCase()) {
          // Remove old token first then re-add it to incorporate any modifications
          const idx = this.customTokens.findIndex(element => element.currencyCode === ethTokenObj.currencyCode);
          if (idx !== -1) {
            this.customTokens.splice(idx, 1);
          }
        }
      }

      // Create a token object for inclusion in customTokens
      const denom = {
        name: ethTokenObj.currencyCode,
        multiplier: ethTokenObj.multiplier
      };
      const edgeMetaToken = {
        currencyCode: ethTokenObj.currencyCode,
        currencyName: ethTokenObj.currencyName,
        denominations: [denom],
        contractAddress
      };

      this.customTokens.push(edgeMetaToken);
      this.allTokens = this.currencyInfo.metaTokens.concat(this.customTokens);
      this.enableTokensSync([edgeMetaToken.currencyCode]);
    } else {
      throw new Error('Invalid custom token object');
    }
  }

  // synchronous
  getTokenStatus(token) {
    return this.walletLocalData.enabledTokens.indexOf(token) !== -1;
  }

  // synchronous
  getBalance(options) {
    let currencyCode = PRIMARY_CURRENCY;

    if (typeof options !== 'undefined') {
      const valid = validateObject(options, {
        type: 'object',
        properties: {
          currencyCode: { type: 'string' }
        }
      });

      if (valid) {
        currencyCode = options.currencyCode;
      }
    }

    if (typeof this.walletLocalData.totalBalances[currencyCode] === 'undefined') {
      return '0';
    } else {
      const nativeBalance = this.walletLocalData.totalBalances[currencyCode];
      return nativeBalance;
    }
  }

  // synchronous
  getNumTransactions(options) {
    let currencyCode = PRIMARY_CURRENCY;

    const valid = validateObject(options, {
      type: 'object',
      properties: {
        currencyCode: { type: 'string' }
      }
    });

    if (valid) {
      currencyCode = options.currencyCode;
    }

    if (typeof this.walletLocalData.transactionsObj[currencyCode] === 'undefined') {
      return 0;
    } else {
      return this.walletLocalData.transactionsObj[currencyCode].length;
    }
  }

  // asynchronous
  async getTransactions(options) {
    let currencyCode = PRIMARY_CURRENCY;

    const valid = validateObject(options, {
      type: 'object',
      properties: {
        currencyCode: { type: 'string' }
      }
    });

    if (valid) {
      currencyCode = options.currencyCode;
    }

    if (typeof this.walletLocalData.transactionsObj[currencyCode] === 'undefined') {
      return [];
    }

    let startIndex = 0;
    let numEntries = 0;
    if (options === null) {
      return this.walletLocalData.transactionsObj[currencyCode].slice(0);
    }
    if (options.startIndex !== null && options.startIndex > 0) {
      startIndex = options.startIndex;
      if (startIndex >= this.walletLocalData.transactionsObj[currencyCode].length) {
        startIndex = this.walletLocalData.transactionsObj[currencyCode].length - 1;
      }
    }
    if (options.numEntries !== null && options.numEntries > 0) {
      numEntries = options.numEntries;
      if (numEntries + startIndex > this.walletLocalData.transactionsObj[currencyCode].length) {
        // Don't read past the end of the transactionsObj
        numEntries = this.walletLocalData.transactionsObj[currencyCode].length - startIndex;
      }
    }

    // Copy the appropriate entries from the arrayTransactions
    let returnArray = [];

    if (numEntries) {
      returnArray = this.walletLocalData.transactionsObj[currencyCode].slice(startIndex, numEntries + startIndex);
    } else {
      returnArray = this.walletLocalData.transactionsObj[currencyCode].slice(startIndex);
    }
    return returnArray;
  }

  // synchronous
  getFreshAddress(options) {
    return { publicAddress: this.walletLocalData.ethereumAddress };
  }

  // synchronous
  addGapLimitAddresses(addresses, options) {}

  // synchronous
  isAddressUsed(address, options) {
    return false;
  }

  // synchronous
  async makeSpend(edgeSpendInfo) {
    // Validate the spendInfo
    const valid = validateObject(edgeSpendInfo, {
      type: 'object',
      properties: {
        currencyCode: { type: 'string' },
        networkFeeOption: { type: 'string' },
        spendTargets: {
          type: 'array',
          items: {
            type: 'object',
            properties: {
              currencyCode: { type: 'string' },
              publicAddress: { type: 'string' },
              nativeAmount: { type: 'string' },
              destMetadata: { type: 'object' },
              destWallet: { type: 'object' }
            },
            required: ['publicAddress']
          }
        }
      },
      required: ['spendTargets']
    });

    if (!valid) {
      throw new Error('Error: invalid ABCSpendInfo');
    }

    // Ethereum can only have one output
    if (edgeSpendInfo.spendTargets.length !== 1) {
      throw new Error('Error: only one output allowed');
    }

    let tokenInfo = {};
    tokenInfo.contractAddress = '';

    let currencyCode = '';
    if (typeof edgeSpendInfo.currencyCode === 'string') {
      currencyCode = edgeSpendInfo.currencyCode;
      if (!this.getTokenStatus(currencyCode)) {
        throw new Error('Error: Token not supported or enabled');
      } else if (currencyCode !== 'ETH') {
        tokenInfo = this.getTokenInfo(currencyCode);
        if (!tokenInfo || typeof tokenInfo.contractAddress !== 'string') {
          throw new Error('Error: Token not supported or invalid contract address');
        }
      }
    } else {
      currencyCode = 'ETH';
    }
    edgeSpendInfo.currencyCode = currencyCode;

    // ******************************
    // Get the fee amount

    let ethParams = {};
    const { gasLimit, gasPrice } = calcMiningFee(edgeSpendInfo, this.walletLocalData.networkFees);

    let publicAddress = '';
    if (typeof edgeSpendInfo.spendTargets[0].publicAddress === 'string') {
      publicAddress = edgeSpendInfo.spendTargets[0].publicAddress;
    } else {
      throw new Error('No valid spendTarget');
    }

    if (currencyCode === PRIMARY_CURRENCY) {
      ethParams = new EthereumParams([this.walletLocalData.ethereumAddress], [publicAddress], gasLimit, gasPrice, '0', '0', 0, null);
    } else {
      let contractAddress = '';
      if (typeof tokenInfo.contractAddress === 'string') {
        contractAddress = tokenInfo.contractAddress;
      } else {
        throw new Error('makeSpend: Invalid contract address');
      }
      ethParams = new EthereumParams([this.walletLocalData.ethereumAddress], [contractAddress], gasLimit, gasPrice, '0', '0', 0, publicAddress);
    }

    let nativeAmount = '0';
    if (typeof edgeSpendInfo.spendTargets[0].nativeAmount === 'string') {
      nativeAmount = edgeSpendInfo.spendTargets[0].nativeAmount;
    } else {
      throw new Error('Error: no amount specified');
    }

    const InsufficientFundsError = new Error('Insufficient funds');
    InsufficientFundsError.name = 'ErrorInsufficientFunds';
    const InsufficientFundsEthError = new Error('Insufficient ETH for transaction fee');
    InsufficientFundsEthError.name = 'ErrorInsufficientFundsMoreEth';

    // Check for insufficient funds
    // let nativeAmountBN = new BN(nativeAmount, 10)
    // const gasPriceBN = new BN(gasPrice, 10)
    // const gasLimitBN = new BN(gasLimit, 10)
    // const nativeNetworkFeeBN = gasPriceBN.mul(gasLimitBN)
    // const balanceEthBN = new BN(this.walletLocalData.totalBalances.ETH, 10)

    const balanceEth = this.walletLocalData.totalBalances.ETH;
    let nativeNetworkFee = biggystring.bns.mul(gasPrice, gasLimit);
    let totalTxAmount = '0';
    let parentNetworkFee = null;

    if (currencyCode === PRIMARY_CURRENCY) {
      totalTxAmount = biggystring.bns.add(nativeNetworkFee, nativeAmount);
      if (biggystring.bns.gt(totalTxAmount, balanceEth)) {
        throw InsufficientFundsError;
      }
      nativeAmount = biggystring.bns.mul(totalTxAmount, '-1');
    } else {
      parentNetworkFee = nativeNetworkFee;

      if (biggystring.bns.gt(nativeNetworkFee, balanceEth)) {
        throw InsufficientFundsEthError;
      }

      nativeNetworkFee = '0'; // Do not show a fee for token transations.
      const balanceToken = this.walletLocalData.totalBalances[currencyCode];
      if (biggystring.bns.gt(nativeAmount, balanceToken)) {
        throw InsufficientFundsError;
      }
      nativeAmount = biggystring.bns.mul(nativeAmount, '-1');
    }

    // const negativeOneBN = new BN('-1', 10)
    // nativeAmountBN.imul(negativeOneBN)
    // nativeAmount = nativeAmountBN.toString(10)

    // **********************************
    // Create the unsigned EdgeTransaction

    const edgeTransaction = {
      txid: '', // txid
      date: 0, // date
      currencyCode, // currencyCode
      blockHeight: 0, // blockHeight
      nativeAmount, // nativeAmount
      networkFee: nativeNetworkFee, // networkFee
      ourReceiveAddresses: [], // ourReceiveAddresses
      signedTx: '0', // signedTx
      otherParams: ethParams // otherParams
    };

    if (parentNetworkFee) {
      edgeTransaction.parentNetworkFee = parentNetworkFee;
    }

    return edgeTransaction;
  }

  // asynchronous
  async signTx(edgeTransaction) {
    // Do signing

    const gasLimitHex = toHex(edgeTransaction.otherParams.gas);
    const gasPriceHex = toHex(edgeTransaction.otherParams.gasPrice);
    let nativeAmountHex;

    // let nativeAmountHex = bns.mul('-1', edgeTransaction.nativeAmount, 16)
    if (edgeTransaction.currencyCode === PRIMARY_CURRENCY) {
      // Remove the networkFee from the nativeAmount
      const nativeAmount = biggystring.bns.add(edgeTransaction.nativeAmount, edgeTransaction.networkFee);
      nativeAmountHex = biggystring.bns.mul('-1', nativeAmount, 16);
    } else {
      nativeAmountHex = biggystring.bns.mul('-1', edgeTransaction.nativeAmount, 16);
    }

    // const nonceBN = new BN(this.walletLocalData.nextNonce.toString(10), 10)
    // const nonceHex = '0x' + nonceBN.toString(16)
    //
    const nonceHex = toHex(this.walletLocalData.nextNonce);
    let data;
    if (edgeTransaction.currencyCode === PRIMARY_CURRENCY) {
      data = '';
    } else {
      const dataArray = abi.simpleEncode('transfer(address,uint256):(uint256)', edgeTransaction.otherParams.tokenRecipientAddress, nativeAmountHex);
      data = '0x' + Buffer$1.from(dataArray).toString('hex');
      nativeAmountHex = '0x00';
    }

    const txParams = {
      nonce: nonceHex,
      gasPrice: gasPriceHex,
      gasLimit: gasLimitHex,
      to: edgeTransaction.otherParams.to[0],
      value: nativeAmountHex,
      data: data,
      // EIP 155 chainId - mainnet: 1, ropsten: 3
      chainId: 1
    };

    const privKey = Buffer$1.from(this.walletInfo.keys.ethereumKey, 'hex');
    const wallet = ethWallet$1.fromPrivateKey(privKey);

    this.log(wallet.getAddressString());

    const tx = new EthereumTx(txParams);
    tx.sign(privKey);

    edgeTransaction.signedTx = bufToHex(tx.serialize());
    edgeTransaction.txid = bufToHex(tx.hash());
    edgeTransaction.date = Date.now() / 1000;

    return edgeTransaction;
  }

  async broadcastEtherscan(edgeTransaction) {
    const result = {
      incrementNonce: false,
      decrementNonce: false
    };
    const transactionParsed = JSON.stringify(edgeTransaction, null, 2);

    this.log(`Etherscan: sent transaction to network:\n${transactionParsed}\n`);
    const url = sprintfJs.sprintf('?module=proxy&action=eth_sendRawTransaction&hex=%s', edgeTransaction.signedTx);
    const jsonObj = await this.fetchGetEtherscan(url);

    this.log('broadcastEtherscan jsonObj:', jsonObj);

    if (typeof jsonObj.error !== 'undefined') {
      this.log('Error sending transaction');
      if (jsonObj.error.code === -32000 || jsonObj.error.message.includes('nonce is too low') || jsonObj.error.message.includes('nonce too low') || jsonObj.error.message.includes('incrementing the nonce') || jsonObj.error.message.includes('replacement transaction underpriced')) {
        result.incrementNonce = true;
      } else {
        throw jsonObj.error;
      }
      return result;
    } else if (typeof jsonObj.result === 'string') {
      // Success!!
      return result;
    } else {
      throw new Error('Invalid return value on transaction send');
    }
  }

  async broadcastBlockCypher(edgeTransaction) {
    const result = {
      incrementNonce: false,
      decrementNonce: false
    };

    const transactionParsed = JSON.stringify(edgeTransaction, null, 2);
    this.log(`Blockcypher: sent transaction to network:\n${transactionParsed}\n`);

    const url = sprintfJs.sprintf('v1/eth/main/txs/push');
    const hexTx = edgeTransaction.signedTx.replace('0x', '');
    const jsonObj = await this.fetchPostBlockcypher(url, { tx: hexTx });

    this.log('broadcastBlockCypher jsonObj:', jsonObj);
    if (typeof jsonObj.error !== 'undefined') {
      this.log('Error sending transaction');
      if (typeof jsonObj.error === 'string' && jsonObj.error.includes('Account nonce ') && jsonObj.error.includes('higher than transaction')) {
        result.incrementNonce = true;
      } else if (typeof jsonObj.error === 'string' && jsonObj.error.includes('Error validating transaction') && jsonObj.error.includes('orphaned, missing reference')) {
        result.decrementNonce = true;
      } else {
        throw jsonObj.error;
      }
      return result;
    } else if (jsonObj.tx && typeof jsonObj.tx.hash === 'string') {
      // Success!!
      return result;
    } else {
      throw new Error('Invalid return value on transaction send');
    }
  }

  // asynchronous
  async broadcastTx(edgeTransaction) {
    const results = [null, null];
    const errors = [null, null];

    // Because etherscan will allow use of a nonce that's too high, only use it if Blockcypher fails
    // If we can fix this or replace etherscan, then we can use an array of promises instead of await
    // on each broadcast type
    try {
      results[0] = await this.broadcastBlockCypher(edgeTransaction);
    } catch (e) {
      errors[0] = e;
    }

    if (errors[0]) {
      try {
        results[1] = await this.broadcastEtherscan(edgeTransaction);
      } catch (e) {
        errors[1] = e;
      }
    }

    // Use code below once we actually use a Promise array and simultaneously broadcast with a Promise.all()
    //
    // for (let i = 0; i < results.length; i++) {
    //   results[i] = null
    //   errors[i] = null
    //   try {
    //     results[i] = await results[i]
    //   } catch (e) {
    //     errors[i] = e
    //   }
    // }

    let allErrored = true;

    for (const e of errors) {
      if (!e) {
        allErrored = false;
        break;
      }
    }

    let anyResultIncNonce = false;
    let anyResultDecrementNonce = false;

    for (const r of results) {
      if (r && r.incrementNonce) {
        anyResultIncNonce = true;
      }
      if (r && r.decrementNonce) {
        anyResultDecrementNonce = true;
      }
    }

    if (allErrored) {
      throw errors[0]; // Can only throw one error so throw the first one
    }

    this.log('broadcastTx errors:', errors);
    this.log('broadcastTx results:', results);

    if (anyResultDecrementNonce) {
      this.walletLocalData.nextNonce = biggystring.bns.add(this.walletLocalData.nextNonce, '-1');
      this.log('Nonce too high. Decrementing to ' + this.walletLocalData.nextNonce.toString());
      // Nonce error. Increment nonce and try again
      const edgeTx = await this.signTx(edgeTransaction);
      const out = await this.broadcastTx(edgeTx);
      return out;
    }

    if (anyResultIncNonce) {
      // All servers returned a nonce-too-low. Increment and retry sign and broadcast
      this.walletLocalData.nextNonce = biggystring.bns.add(this.walletLocalData.nextNonce, '1');
      this.log('Nonce too low. Incrementing to ' + this.walletLocalData.nextNonce.toString());
      // Nonce error. Increment nonce and try again
      const edgeTx = await this.signTx(edgeTransaction);
      const out = await this.broadcastTx(edgeTx);
      return out;
    }
    // Success
    this.walletLocalData.nextNonce = biggystring.bns.add(this.walletLocalData.nextNonce, '1');

    return edgeTransaction;
  }

  // asynchronous
  async saveTx(edgeTransaction) {
    this.addTransaction(edgeTransaction.currencyCode, edgeTransaction);

    this.edgeTxLibCallbacks.onTransactionsChanged([edgeTransaction]);
  }

  getDisplayPrivateSeed() {
    if (this.walletInfo.keys && this.walletInfo.keys.ethereumKey) {
      return this.walletInfo.keys.ethereumKey;
    }
    return '';
  }

  getDisplayPublicSeed() {
    if (this.walletInfo.keys && this.walletInfo.keys.ethereumAddress) {
      return this.walletInfo.keys.ethereumAddress;
    }
    return '';
  }

  dumpData() {
    const dataDump = {
      walletId: this.walletId.split(' - ')[0],
      walletType: this.walletInfo.type,
      pluginType: this.currencyInfo.pluginName,
      data: {
        walletLocalData: this.walletLocalData
      }
    };
    return dataDump;
  }
}

/**
 * Created by paul on 8/8/17.
 */
const Buffer = require('buffer/').Buffer;
const ethWallet = require('../lib/export-fixes-bundle.js').Wallet;
const EthereumUtil = require('../lib/export-fixes-bundle.js').Util;

let io;

const randomBuffer = size => {
  const array = io.random(size);
  return Buffer.from(array);
};

function getDenomInfo(denom) {
  return currencyInfo.denominations.find(element => {
    return element.name === denom;
  });
}

function hexToBuf(hex) {
  const noHexPrefix = hex.replace('0x', '');
  const noHexPrefixBN = new bn_js.BN(noHexPrefix, 16);
  const array = noHexPrefixBN.toArray();
  const buf = Buffer.from(array);
  return buf;
}

function getParameterByName(param, url) {
  const name = param.replace(/[[\]]/g, '\\$&');
  const regex = new RegExp('[?&]' + name + '(=([^&#]*)|&|#|$)');
  const results = regex.exec(url);
  if (!results) return null;
  if (!results[2]) return '';
  return decodeURIComponent(results[2].replace(/\+/g, ' '));
}

// async function checkUpdateCurrencyInfo () {
//   while (this.engineOn) {
//     try {
//       const url = sprintf('%s/v1/currencyInfo/ETH', INFO_SERVERS[0])
//       const jsonObj = await this.fetchGet(url)
//       const valid = validateObject(jsonObj, CurrencyInfoScheme)
//
//       if (valid) {
//         console.log('Fetched valid currencyInfo')
//         console.log(jsonObj)
//       } else {
//         console.log('Error: Fetched invalid currencyInfo')
//       }
//     } catch (err) {
//       console.log('Error fetching currencyInfo: ' + err)
//     }
//     try {
//       await snooze(BLOCKHEIGHT_POLL_MILLISECONDS)
//     } catch (err) {
//       console.log(err)
//     }
//   }
// }

const ethereumCurrencyPluginFactory = {
  pluginType: 'currency',
  pluginName: currencyInfo.pluginName,

  async makePlugin(opts) {
    io = opts.io;

    console.log(`Creating Currency Plugin for ethereum`);
    const ethereumPlugin = {
      pluginName: 'ethereum',
      currencyInfo,

      createPrivateKey: walletType => {
        const type = walletType.replace('wallet:', '');

        if (type === 'ethereum') {
          const cryptoObj = {
            randomBytes: randomBuffer
          };
          ethWallet.overrideCrypto(cryptoObj);

          const wallet = ethWallet.generate(false);
          const ethereumKey = wallet.getPrivateKeyString().replace('0x', '');
          return { ethereumKey };
        } else {
          throw new Error('InvalidWalletType');
        }
      },

      derivePublicKey: walletInfo => {
        const type = walletInfo.type.replace('wallet:', '');
        if (type === 'ethereum') {
          const privKey = hexToBuf(walletInfo.keys.ethereumKey);
          const wallet = ethWallet.fromPrivateKey(privKey);

          const ethereumAddress = wallet.getAddressString();
          // const ethereumKey = '0x389b07b3466eed587d6bdae09a3613611de9add2635432d6cd1521af7bbc3757'
          // const ethereumPublicAddress = '0x9fa817e5A48DD1adcA7BEc59aa6E3B1F5C4BeA9a'
          return { ethereumAddress };
        } else {
          throw new Error('InvalidWalletType');
        }
      },

      // XXX Deprecated. To be removed once Core supports createPrivateKey and derivePublicKey -paulvp
      createMasterKeys: walletType => {
        if (walletType === 'ethereum') {
          const cryptoObj = {
            randomBytes: randomBuffer
          };
          ethWallet.overrideCrypto(cryptoObj);

          const wallet = ethWallet.generate(false);
          const ethereumKey = wallet.getPrivateKeyString().replace('0x', '');
          const ethereumPublicAddress = wallet.getAddressString();
          // const ethereumKey = '0x389b07b3466eed587d6bdae09a3613611de9add2635432d6cd1521af7bbc3757'
          // const ethereumPublicAddress = '0x9fa817e5A48DD1adcA7BEc59aa6E3B1F5C4BeA9a'
          return { ethereumKey, ethereumPublicAddress };
        } else {
          return null;
        }
      },

      async makeEngine(walletInfo, opts) {
        const ethereumEngine = new EthereumEngine(io, walletInfo, opts);
        try {
          const result = await ethereumEngine.walletLocalFolder.folder(DATA_STORE_FOLDER).file(DATA_STORE_FILE).getText(DATA_STORE_FOLDER, 'walletLocalData');

          ethereumEngine.walletLocalData = new WalletLocalData(result);
          ethereumEngine.walletLocalData.ethereumAddress = ethereumEngine.walletInfo.keys.ethereumAddress;
        } catch (err) {
          try {
            console.log(err);
            console.log('No walletLocalData setup yet: Failure is ok');
            ethereumEngine.walletLocalData = new WalletLocalData(null);
            ethereumEngine.walletLocalData.ethereumAddress = ethereumEngine.walletInfo.keys.ethereumAddress;
            await ethereumEngine.walletLocalFolder.folder(DATA_STORE_FOLDER).file(DATA_STORE_FILE).setText(JSON.stringify(ethereumEngine.walletLocalData));
          } catch (e) {
            console.log('Error writing to localDataStore. Engine not started:' + err);
          }
        }
        for (const token of ethereumEngine.walletLocalData.enabledTokens) {
          ethereumEngine.tokenCheckStatus[token] = 0;
        }
        return ethereumEngine;
      },

      parseUri: uri => {
        const parsedUri = uriJs.parse(uri);
        let address;
        let nativeAmount = null;
        let currencyCode = null;

        if (typeof parsedUri.scheme !== 'undefined' && parsedUri.scheme !== 'ethereum' && parsedUri.scheme !== 'ether') {
          throw new Error('InvalidUriError'); // possibly scanning wrong crypto type
        }
        if (typeof parsedUri.host !== 'undefined') {
          address = parsedUri.host;
        } else if (typeof parsedUri.path !== 'undefined') {
          address = parsedUri.path;
        } else {
          throw new Error('InvalidUriError');
        }
        address = address.replace('/', ''); // Remove any slashes
        let [prefix, contractAddress] = address.split('-'); // Split the address to get the prefix according to EIP-681
        // If contractAddress is null or undefined it means there is no prefix
        if (!contractAddress) {
          contractAddress = prefix; // Set the contractAddress to be the prefix when the prefix is missing.
          prefix = 'pay'; // The default prefix according to EIP-681 is "pay"
        }
        address = contractAddress;
        const valid = EthereumUtil.isValidAddress(address);
        if (!valid) {
          throw new Error('InvalidPublicAddressError');
        }
        // If the address has a "token-" prefix, it means it's an "Add Token" URI and not a payment one.
        if (prefix === 'token' || prefix === 'token_info') {
          const currencyCode = getParameterByName('symbol', uri) || 'SYM';
          if (currencyCode.length < 2 || currencyCode.length > 5) {
            throw new Error('Wrong Token symbol');
          }
          const currencyName = getParameterByName('name', uri) || currencyCode;
          const decimalsInput = getParameterByName('decimals', uri) || '18';
          let multiplier = '1000000000000000000';
          try {
            const decimals = parseInt(decimalsInput);
            if (decimals < 0 || decimals > 18) {
              throw new Error('Wrong number of decimals');
            }
            multiplier = '1' + '0'.repeat(decimals);
          } catch (e) {
            throw e;
          }

          const type = getParameterByName('type', uri) || 'ERC20';

          const edgeParsedUri = {
            token: {
              currencyCode,
              contractAddress,
              currencyName,
              multiplier,
              type: type.toUpperCase()
            }
          };
          return edgeParsedUri;
        }
        const amountStr = getParameterByName('amount', uri);
        if (amountStr && typeof amountStr === 'string') {
          const denom = getDenomInfo('ETH');
          if (!denom) {
            throw new Error('InternalErrorInvalidCurrencyCode');
          }
          nativeAmount = biggystring.bns.mul(amountStr, denom.multiplier);
          nativeAmount = biggystring.bns.toFixed(nativeAmount, 0, 0);
          currencyCode = 'ETH';
        }
        const label = getParameterByName('label', uri);
        const message = getParameterByName('message', uri);

        const edgeParsedUri = {
          publicAddress: address
        };
        if (nativeAmount) {
          edgeParsedUri.nativeAmount = nativeAmount;
        }
        if (currencyCode) {
          edgeParsedUri.currencyCode = currencyCode;
        }
        if (label || message) {
          edgeParsedUri.metadata = {};
          if (label) {
            edgeParsedUri.metadata.name = label;
          }
          if (message) {
            edgeParsedUri.metadata.message = message;
          }
        }

        return edgeParsedUri;
      },

      encodeUri: obj => {
        if (!obj.publicAddress) {
          throw new Error('InvalidPublicAddressError');
        }
        const valid = EthereumUtil.isValidAddress(obj.publicAddress);
        if (!valid) {
          throw new Error('InvalidPublicAddressError');
        }
        if (!obj.nativeAmount && !obj.label && !obj.message) {
          return obj.publicAddress;
        } else {
          let queryString = '';
          const hack = obj; // EdgeEncodeUri does not have a currencyCode property
          if (typeof obj.nativeAmount === 'string') {
            let currencyCode = 'ETH';
            const nativeAmount = obj.nativeAmount;

            if (typeof hack.currencyCode === 'string') {
              currencyCode = hack.currencyCode;
            }
            const denom = getDenomInfo(currencyCode);
            if (!denom) {
              throw new Error('InternalErrorInvalidCurrencyCode');
            }
            const amount = biggystring.bns.div(nativeAmount, denom.multiplier, 18);

            queryString += 'amount=' + amount + '&';
          }
          if (hack.metadata && (hack.metadata.name || hack.metadata.message)) {
            if (typeof hack.metadata.name === 'string') {
              queryString += 'label=' + hack.metadata.name + '&';
            }
            if (typeof hack.metadata.message === 'string') {
              queryString += 'message=' + hack.metadata.message + '&';
            }
          }
          queryString = queryString.substr(0, queryString.length - 1);

          const serializeObj = {
            scheme: 'ethereum',
            path: obj.publicAddress,
            query: queryString
          };
          const url = uriJs.serialize(serializeObj);
          return url;
        }
      }
    };

    async function initPlugin(opts) {
      // Try to grab currencyInfo from disk. If that fails, use defaults

      // try {
      //   const result =
      //     await this.walletLocalFolder
      //       .folder(DATA_STORE_FOLDER)
      //       .file(DATA_STORE_FILE)
      //       .getText(DATA_STORE_FOLDER, 'walletLocalData')
      //
      //   this.walletLocalData = new WalletLocalData(result)
      //   this.walletLocalData.ethereumAddress = this.walletInfo.keys.ethereumAddress
      // }

      // Spin off network query to get updated currencyInfo and save that to disk for future bootups

      return ethereumPlugin;
    }
    return initPlugin(opts);
  }
};

exports.ethereumCurrencyPluginFactory = ethereumCurrencyPluginFactory;
exports.EthereumCurrencyPluginFactory = ethereumCurrencyPluginFactory;
exports.calcMiningFee = calcMiningFee;
//# sourceMappingURL=indexEthereum.js.map
